/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#include <Controller.h>
#include "DualSwitchSpectroController.h"
#include "controllers/common/family/Families.h"
#include "InstrumentAbstraction/Base/AbstractIOController.h"
#include "Utilities/Counter.h"
#include "ics/DeviceContainer/Def/IODef.h"

namespace d22special {

const std::string DualSwitchSpectroController::TYPE = "dualSwitchSpectro";

DualSwitchSpectroController::DualSwitchSpectroController(const std::string& name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::SAMPLE_ENVIRONMENT);

	//init properties
	integrationTime.init(this, SAVE, "integration_time");
	numberOfCycles.init(this, SAVE, "number_cycles");
	executionStatus.init(this, SAVE, "execution_status");

	xData.init(this, SAVE, "x_data");
	xLeftData.init(this, SAVE, "x_left_data");
	yLeftData.init(this, SAVE, "y_leftData");
	yRightData.init(this, SAVE, "y_rightData");

	m_StopActivated = false;

	//It has attached one module (QE65000 controller) and one controller (switcher)
	spectroController.init(this, "spectro");
	switcherController.init(this, "swicther");

}

DualSwitchSpectroController::~DualSwitchSpectroController() {

}

void DualSwitchSpectroController::postConfiguration() {
	registerProgression(spectroController, &DualSwitchSpectroController::updateProgression, this);
}

void DualSwitchSpectroController::start() {

	m_StopActivated = false;
	commandProgression = 0;
//spectroController->mode = QE65000Controller::SWITCH;

	//Before measuring spectrum check if the spectrometer is plugged in
//	if (spectroController->isConnected()) {
	executionStatus.update("idle");

	//Set up values of spectro
	spectroController->integrationTime = integrationTime();

	int32 cyclesNumber = numberOfCycles();
	while (cyclesNumber > 0 && !m_StopActivated) {

		commandProgression = (numberOfCycles() - cyclesNumber) * 100 / numberOfCycles();

		// First: start switcher -> we begin for OPEN position (== IN)
		executionStatus.update("Measuring Left");
		switcherController->setOpen();
		switcherController->startCommand();

		//Second: start spectro
		spectroController->start();
		updateData();

		//Again same process: start switcher (now to CLOSE == OUT) and then spectro
		executionStatus.update("Measuring Right");
		switcherController->setClose();
		switcherController->startCommand();

		spectroController->start();
		updateData();

		cyclesNumber = cyclesNumber - 1;

	}
	commandProgression = 99;
	executionStatus.update("finished");

//	} else {
//		cout << "spectro disconnected" << endl;
//		executionStatus.update("disconnected");
//	}
}

void DualSwitchSpectroController::stop() {
	m_StopActivated = true;
	commandProgression = 100;
}

void DualSwitchSpectroController::updateProgression() {
//	int32 progression = 0;
//	commandProgression = progression;
}

void DualSwitchSpectroController::updateData() {
	//should be done with an updater...
	if (executionStatus() == "Measuring Right") {
		int32 xSize = spectroController->xData.getSize();
		xData.update(spectroController->xData());
		xData.setSize(xSize);
		xData.sendEvent();

		int32 yRightSize = spectroController->yData.getSize();
		yRightData.update(spectroController->yData());
		yRightData.setSize(yRightSize);
		yRightData.sendEvent();

	} else {
		//testing two x arrays of data
		int32 xSize = spectroController->xData.getSize();
		xLeftData.update(spectroController->xData());
		xLeftData.setSize(xSize);
		xLeftData.sendEvent();

		int32 yLeftSize = spectroController->yData.getSize();
		yLeftData.update(spectroController->yData());
		yLeftData.setSize(yLeftSize);
		yLeftData.sendEvent();
	}

}

}

