/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteDPPHistogram.h"
#include <common/base/ServerProperties.h>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <cstdlib>

namespace npp {

using namespace std;
using namespace common;
using namespace boost;
using namespace nappli;

const std::string RemoteDPPHistogram::TYPE = "remote_dpp_histogram";

RemoteDPPHistogram::RemoteDPPHistogram(const std::string& name) :
	DPPLiveProcess(name) {

	serverEndpoint.init(this, SAVE, "nappli_server");
	triggerChannel.init(this, SAVE, "trigger_channel");
	delay.init(this, SAVE, "delay");
	dT.init(this, SAVE, "dT");
	maxBlock.init(this, SAVE, "max_block");
	blockSize.init(this, SAVE, "block_size");
	test.init(this, SAVE, "test");
	testFileName.init(this, SAVE, "test_file_name");
	xData.init(this, NOSAVE, "x_data");
	one.init(this, NOSAVE, "one");
	refreshTimeS.init(this, SAVE, "refresh_time_s");
	histogram.init(this, NOSAVE, "histogram");
	ADCId.init(this, NOSAVE, "adc_id");
	ADCRate.init(this, NOSAVE, "adc_rate");
}

RemoteDPPHistogram::RemoteDPPHistogram(const RemoteDPPHistogram& controller) :
	DPPLiveProcess(controller) {
}

RemoteDPPHistogram::~RemoteDPPHistogram() {
}

void RemoteDPPHistogram::postConfiguration() {

	string endpoint = serverEndpoint();
	if (endpoint == "") {
		m_server = Application::getLocalServer();

	} else {
		m_server.reset(new Server(endpoint));
	}

	histogram.resize(MAX_ADC);
	ADCId.resize(MAX_ADC);
	ADCRate.resize(MAX_ADC);

	// allocating
	for (int32 i = 0; i < MAX_ADC; i++) {
		ADCId.set(i, i);
	}

	resetData();
}

void RemoteDPPHistogram::setDetectorControllers(const std::vector<ControllerPtr<acquisition::DetectorController> > & detectorControllers) {

	DPPLiveProcess::setDetectorControllers(detectorControllers);

	// register the list mode data
	int32 index = 0;
	for (vector<ControllerPtr<acquisition::DetectorController> >::const_iterator c = detectorControllers.begin(); c != detectorControllers.end(); ++c) {
		registerUpdater((*c)->listModeData, &RemoteDPPHistogram::updateListModeData, this, index);
		index++;
	}
}

void RemoteDPPHistogram::start() {

	auto_ptr<DPPReader> reader = initTest();

	resetData();

	// napplihisto <numor> <number of ADCs> <number of channels> <channel> <delay> <dT> <crateboard.txt> [maxBlocks] [blockSize] [result period]
	vector<string> args;

	// numor
	string numor;
	size_t size = fileName().size();

	if (size < 6) {
		string prefix(6 - size, '0');
		numor = prefix + fileName();
	} else {
		numor = fileName();
	}
	args.push_back(numor);
	args.push_back(lexical_cast<string>(numberOfADCs()));
	args.push_back(lexical_cast<string>(numberOfChannels()));
	args.push_back(lexical_cast<string>(triggerChannel()));
	args.push_back(lexical_cast<string>(delay() / 10));
	args.push_back(lexical_cast<string>(dT() / 10));
	args.push_back(ServerProperties::getInstance()->getHomePath() + "/.nomadserver/crateboardhistogram.txt");
	args.push_back(lexical_cast<string>(maxBlock()));
	args.push_back(lexical_cast<string>(blockSize()));
	args.push_back(lexical_cast<string>(refreshTimeS()));

	// kill the application if one remains
	m_server->killAllAndWaitFor("nhisto");

	// start the application
	m_histoApplication = m_server->start("nhisto", args);

	// test if application started
	if (!m_histoApplication->exists()) {
		log(Level::s_Error) << "cannot start histogram application" << endlog;
		return;
	}

	// create the publisher after the start of the application, because it is a blocking call
	m_publisher = Application::publish("dpp_publisher", 1);

	cout << "publisher " << *m_publisher << endl;

	// create the subscriber after the publisher
	auto_ptr<nappli::Application::Subscriber> subscriber = m_histoApplication->subscribe("dpp_histogram_results");
	cout << "subscriber " << *subscriber << endl;

	// start the test thread if necessary
	auto_ptr<boost::thread> thread = startTest(reader.get());

	// listen the result stream
	vector<double> ADCRates;
	while (subscriber->receive(ADCRates)) {
		// copying the rates
		for (int32 i = 0; i < numberOfADCs(); i++) {
			ADCRate.set(i, ADCRates[i]);
		}

		vector<int32_t> histograms;
		if (!subscriber->receive(histograms)) {
			break;
		}

		if (histograms.size() <= numberOfADCs() * numberOfChannels()) {

			int32 N = numberOfADCs();
			int32 M = numberOfChannels();

			for (int32 i = 0; i < N; i++) {

				int32 * histo = histogram(i);

				for (int32 j = 0; j < M; j++) {
					histo[j] = histograms[i * M + j];
				}

				histogram.sendEvent(i);
			}

		} else {
			cerr << "problem with histogram size" << endl;
		}
	}

	// wait for the termination of the application
	Application::State state = m_histoApplication->waitFor();

	cout << "histogram application terminated with state " << state << endl;

	// join the test thread
	finishTest(thread);

	// terminate the publisher
	m_publisher->terminate();
}

void RemoteDPPHistogram::stop() {

	if (m_histoApplication.get() != 0) {
		cout << *m_publisher << " end of stream" << endl;
		m_publisher->setEndOfStream();
		// we do not stop the application because the termination of the publisher will do it.
	}
}

void RemoteDPPHistogram::raz() {
	resetData();
}

void RemoteDPPHistogram::updateListModeData(int index) {

	if (m_publisher.get() != 0 && !m_publisher->isEndOfStream()) {
		int32 size = detectorControllers[index]->listModeData.getSize();
		int32 * memoryBlock = detectorControllers[index]->listModeData();

		// put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32) * size);

		// send the data
		m_publisher->send(data);

		cout << "updateListModeData " << index << " size " << size << endl;
	}
}

auto_ptr<DPPReader> RemoteDPPHistogram::initTest() {

	if (test()) {
		auto_ptr<DPPReader> reader(new DPPReader(blockSize()));

		reader->openFile(testFileName());

		cout << "#ADCs " << reader->numberOfADCs() << endl;
		cout << "#channels " << reader->numberOfChannels() << endl;

		// override real values
		numberOfADCs = reader->numberOfADCs();
		numberOfChannels = reader->numberOfChannels();

		return reader;
	}

	return auto_ptr<DPPReader>();
}

void RemoteDPPHistogram::testLoop(DPPReader * reader) {

	// loop over events
	while (!m_publisher->isEndOfStream()) {
		// reading next block
		int32_t blockSize = reader->readBlock();

		// exiting if no data
		if (blockSize == 0) {
			break;
		}

		// sending block
		int32_t * memoryBlock = reader->memoryBlock();

		// put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32_t) * blockSize);

		// send the data
		m_publisher->send(data);

		//sleep(refreshTimeS());
		usleep(10000);
	}

	cout << "finished reading test file " << testFileName() << endl;
}

std::auto_ptr<boost::thread> RemoteDPPHistogram::startTest(DPPReader * reader) {

	return auto_ptr<thread>(new thread(bind(&RemoteDPPHistogram::testLoop, this, reader)));
}

void RemoteDPPHistogram::finishTest(std::auto_ptr<boost::thread> thread) {

	// join the thread
	thread->join();

	cout << "finished test loop" << endl;
}

void RemoteDPPHistogram::resetData() {

	// number of elements
	int32_t size = numberOfChannels() * numberOfADCs();

	// ADC X data
	delete [] xData();
	xData.set(new int32[numberOfChannels()]);

	// initializing
	for (int32 i = 0; i < numberOfChannels(); ++i) {
		xData()[i] = i;
	}

	xData.setSize(numberOfChannels());

	// setting dynamic properties
	for (int32 i = 0; i < numberOfADCs(); i++) {
		delete [] histogram(i);
		histogram.set(i, new int32[numberOfChannels()]);
		histogram.setSize(i, numberOfChannels());
	}

	for (int32 i = numberOfADCs(); i < MAX_ADC; i++) {
		histogram.set(i, 0);
		histogram.setSize(i, 0);
	}

	for (int32 i = 0; i < MAX_ADC; i++) {
		ADCRate.set(i, 0.0);
	}
}

}
